@@ -109,7 +109,9 @@ if ($config{'system_crontab'}) {
 
 # read package-specific cron files
 opendir(DIR, &translate_filename($config{'cronfiles_dir'}));
-while($f = readdir(DIR)) {
+my @files = sort { $a cmp $b } readdir(DIR);
+closedir(DIR);
+foreach my $f (@files) {
 	next if ($f =~ /^\./);
 	$lnum = 0;
 	&open_readfile(TAB, "$config{'cronfiles_dir'}/$f");
@@ -141,7 +143,6 @@ while($f = readdir(DIR)) {
 		}
 	close(TAB);
 	}
-closedir(DIR);
 
 # Read a single user's crontab file
 if ($config{'single_file'}) {
@@ -171,6 +172,7 @@ if ($config{'single_file'}) {
 				    'name' => $2,
 				    'value' => $3,
 				    'user' => "NONE",
+				    'command' => '',
 				    'index' => scalar(@rv) });
 			}
 		$lnum++;
@@ -280,7 +282,12 @@ else {
 				      $_[0]->{'type'} != 3);
 	push(@c, $_[0]->{'command'});
 	}
-return join(" ", @c);
+if ($gconfig{'os_type'} eq 'syno-linux') {
+	return join("\t", @c);
+	}
+else {
+	return join(" ", @c);
+	}
 }
 
 =head2 copy_cron_temp(&job)
@@ -302,8 +309,9 @@ elsif ($fcron) {
 			 undef, $cron_temp_file, undef);
 	}
 else {
-	system("cp ".&translate_filename("$config{'cron_dir'}/$_[0]->{'user'}").
-	       " $cron_temp_file 2>/dev/null");
+	system("cp ".quotemeta(
+			&translate_filename("$config{'cron_dir'}/$_[0]->{'user'}")).
+	       " ".quotemeta($cron_temp_file)." 2>/dev/null");
 	}
 }
 
@@ -337,7 +345,7 @@ else {
 	local $lref = &read_file_lines($cron_temp_file);
 	$_[0]->{'line'} = scalar(@$lref);
 	push(@$lref, &cron_job_line($_[0]));
-	&flush_file_lines();
+	&flush_file_lines($cron_temp_file);
 	&set_ownership_permissions($_[0]->{'user'}, undef, undef,
 				   $cron_temp_file);
 	&copy_crontab($_[0]->{'user'});
@@ -371,7 +379,7 @@ else {
 	$_[0]->{'line'} = 0;
 	splice(@$lref, 0, 0, &cron_job_line($_[0]));
 	&flush_file_lines();
-	system("chown $_[0]->{'user'} $cron_temp_file");
+	&set_ownership_permissions($_[0]->{'user'}, undef, undef, $cron_temp_file);
 	&copy_crontab($_[0]->{'user'});
 	$_[0]->{'file'} = "$config{'cron_dir'}/$_[0]->{'user'}";
 	$_[0]->{'index'} = scalar(@cron_jobs_cache);
@@ -488,9 +496,15 @@ local($pwd);
 if (&read_file_contents($cron_temp_file) =~ /\S/) {
 	local $temp = &transname();
 	local $rv;
-	if ($config{'cron_edit_command'}) {
+	if (!&has_crontab_cmd()) {
+		# We have no crontab command .. emulate by copying to user file
+		$rv = system("cat ".quotemeta($cron_temp_file).
+			" >".quotemeta("$config{'cron_dir'}/$_[0]")." 2>/dev/null");
+		&set_ownership_permissions($_[0], undef, 0600,
+			"$config{'cron_dir'}/$_[0]");
+		}
+	elsif ($config{'cron_edit_command'}) {
 		# fake being an editor
-		# XXX does not work in translated command mode!
 		local $notemp = &transname();
 		&open_tempfile(NO, ">$notemp");
 		&print_tempfile(NO, "No\n");
@@ -500,34 +514,34 @@ if (&read_file_contents($cron_temp_file) =~ /\S/) {
 		$ENV{"VISUAL"} = $ENV{"EDITOR"} =
 			"$module_root_directory/cron_editor.pl";
 		$ENV{"CRON_EDITOR_COPY"} = $cron_temp_file;
-		system("chown $_[0] $cron_temp_file");
+		system("chown ".quotemeta($_[0])." ".quotemeta($cron_temp_file));
 		local $oldpwd = &get_current_dir();
 		chdir("/");
 		if ($single_user) {
 			$rv = system($config{'cron_user_edit_command'}.
-				     " >$temp 2>&1 <$notemp");
+				     " >".quotemeta($temp)." 2>&1 <".quotemeta($notemp));
 			}
 		else {
 			$rv = system(
 				&user_sub($config{'cron_edit_command'},$_[0]).
-				" >$temp 2>&1 <$notemp");
+				" >".quotemeta($temp)." 2>&1 <".quotemeta($notemp));
 			}
 		unlink($notemp);
 		chdir($oldpwd);
-		}
-	else {
+
+	} else {
 		# use the cron copy command
 		if ($single_user) {
 			$rv = &execute_command(
 				$config{'cron_user_copy_command'},
-				$cron_temp_file, $temp, $temp);
+				quotemeta($cron_temp_file), quotemeta($temp), quotemeta($temp));
 			}
 		else {
 			$rv = &execute_command(
 				&user_sub($config{'cron_copy_command'}, $_[0]),
-				$cron_temp_file, $temp, $temp);
+				quotemeta($cron_temp_file), quotemeta($temp), quotemeta($temp));
 			}
-		}
+	}
 	local $out = &read_file_contents($temp);
 	unlink($temp);
 	if ($rv || $out =~ /error/i) {
@@ -537,14 +551,25 @@ if (&read_file_contents($cron_temp_file) =~ /\S/) {
 	}
 else {
 	# No more cron jobs left, so just delete
-	if ($single_user) {
-		&execute_command($config{'cron_user_delete_command'});
+	if (!&has_crontab_cmd()) {
+		# We have no crontab command .. emulate by deleting user crontab
+		$_[0] || &error("No user given!");
+		&unlink_logged("$config{'cron_dir'}/$_[0]");
 		}
-	else {
-		&execute_command(&user_sub(
-			$config{'cron_delete_command'}, $_[0]));
+	else{
+		if ($single_user) {
+			&execute_command($config{'cron_user_delete_command'});
+			}
+		else {
+			&execute_command(&user_sub(
+				$config{'cron_delete_command'}, $_[0]));
+			}
 		}
 	}
+if (!&has_crontab_cmd()) {
+	# to reload config
+	&kill_byname("crond", "SIGHUP");
+	}
 unlink($cron_temp_file);
 }
 
@@ -576,7 +601,7 @@ sub user_sub
 {
 local($tmp);
 $tmp = $_[0];
-$tmp =~ s/USER/$_[1]/g;
+$tmp =~ s/USER/quotemeta($_[1])/ge;
 return $tmp;
 }
 
@@ -677,7 +702,7 @@ sub save_envs
 {
 local($i, @tab, $line);
 @tab = &read_crontab($_[0]);
-open(TAB, ">$cron_temp_file");
+open(TAB, ">".$cron_temp_file);
 for($i=1; $i<@_; $i+=2) {
 	print TAB "$_[$i]=$_[$i+1]\n";
 	}
@@ -703,8 +728,9 @@ $dir = "$config{'run_parts_dir'}/$dir"
 opendir(DIR, &translate_filename($dir));
 local @rv = readdir(DIR);
 closedir(DIR);
-@rv = grep { !/^\./ } @rv;
+@rv = grep { /^[a-zA-Z0-9\_\-]+$/ } @rv;
 @rv = map { $dir."/".$_ } @rv;
+@rv = grep { -x $_ } @rv;
 return @rv;
 }
 
@@ -716,8 +742,10 @@ like /etc/cron.hourly. Returns undef otherwise.
 =cut
 sub is_run_parts
 {
+local ($cmd) = @_;
 local $rp = $config{'run_parts'};
-return $rp && $_[0] =~ /$rp(.*)\s+(\-\-\S+\s+)*([a-z0-9\.\-\/_]+)(\s*\))?$/i ? $3 : undef;
+$cmd =~ s/\s*#.*$//;
+return $rp && $cmd =~ /$rp(.*)\s+(\-\-\S+\s+)*([a-z0-9\.\-\/_]+)(\s*\))?$/i ? $3 : undef;
 }
 
 =head2 can_edit_user(&access, user)
@@ -750,6 +778,158 @@ else {
 	}
 }
 
+=head2 list_cron_specials()
+
+Returns a list of the names of special cron times, prefixed by an @ in crontab
+
+=cut
+sub list_cron_specials
+{
+return ('hourly', 'daily', 'weekly', 'monthly', 'yearly', 'reboot');
+}
+
+=head2 get_times_input(&job, [nospecial], [width-in-cols], [message])
+
+Returns HTML for selecting the schedule for a cron job, defined by the first
+parameter which must be a hash ref returned by list_cron_jobs. Suitable for
+use inside a ui_table_start/end
+
+=cut
+sub get_times_input
+{
+return &theme_get_times_input(@_) if (defined(&theme_get_times_input));
+my ($job, $nospecial, $width, $msg) = @_;
+$width ||= 2;
+
+# Javascript to disable and enable fields
+my $rv = <<EOF;
+<script>
+function enable_cron_fields(name, form, ena)
+{
+var els = form.elements[name];
+els.disabled = !ena;
+for(i=0; i<els.length; i++) {
+  els[i].disabled = !ena;
+  }
+change_special_mode(form, 0);
+}
+
+function change_special_mode(form, special)
+{
+  if(form.special_def) {
+    form.special_def[0].checked = special;
+    form.special_def[1].checked = !special;
+  }
+}
+</script>
+EOF
+
+if ($config{'vixie_cron'} && (!$nospecial || $job->{'special'})) {
+	# Allow selection of special @ times
+	my $sp = $job->{'special'} eq 'midnight' ? 'daily' :
+		 $job->{'special'} eq 'annually' ? 'yearly' : $job->{'special'};
+	my $specialsel = &ui_select("special", $sp,
+			[ map { [ $_, $text{'edit_special_'.$_} ] }
+			      &list_cron_specials() ],
+			1, 0, 0, 0, "onChange='change_special_mode(form, 1)'");
+	$rv .= &ui_table_row($msg,
+		&ui_radio("special_def", $job->{'special'} ? 1 : 0,
+			  [ [ 1, $text{'edit_special1'}."&nbsp;&nbsp;".$specialsel ],
+			    [ 0, $text{'edit_special0'} ] ]),
+			  $msg ? $width-1 : $width);
+	}
+
+# Section for time selections
+my $table = &ui_columns_start([ $text{'edit_mins'}, $text{'edit_hours'},
+				$text{'edit_days'}, $text{'edit_months'},
+				$text{'edit_weekdays'} ], 100);
+my @mins = (0..59);
+my @hours = (0..23);
+my @days = (1..31);
+my @months = map { $text{"month_$_"}."=".$_ } (1 .. 12);
+my @weekdays = map { $text{"day_$_"}."=".$_ } (0 .. 6);
+my %arrmap = ( 'mins' => \@mins,
+	       'hours' => \@hours,
+	       'days' => \@days,
+	       'months' => \@months,
+	       'weekdays' => \@weekdays );
+my @cols;
+foreach my $arr ("mins", "hours", "days", "months", "weekdays") {
+	# Find out which ones are being used
+	my %inuse;
+	my $min = ($arr =~ /days|months/ ? 1 : 0);
+	my @arrlist = @{$arrmap{$arr}};
+	my $max = $min+scalar(@arrlist)-1;
+	foreach my $w (split(/,/ , $job->{$arr})) {
+		if ($w eq "*") {
+			# all values
+			for($j=$min; $j<=$max; $j++) { $inuse{$j}++; }
+			}
+		elsif ($w =~ /^\*\/(\d+)$/) {
+			# only every Nth
+			for($j=$min; $j<=$max; $j+=$1) { $inuse{$j}++; }
+			}
+		elsif ($w =~ /^(\d+)-(\d+)\/(\d+)$/) {
+			# only every Nth of some range
+			for($j=$1; $j<=$2; $j+=$3) { $inuse{int($j)}++; }
+			}
+		elsif ($w =~ /^(\d+)-(\d+)$/) {
+			# all of some range
+			for($j=$1; $j<=$2; $j++) { $inuse{int($j)}++; }
+			}
+		else {
+			# One value
+			$inuse{int($w)}++;
+			}
+		}
+	if ($job->{$arr} eq "*") {
+		%inuse = ( );
+		}
+
+	# Output selection list
+	my $dis = $arr eq "mins" && $hourly_only;
+	my $col = &ui_radio(
+		    "all_$arr", $job->{$arr} eq "*" ||
+				$job->{$arr} eq "" ? 1 : 0,
+		    [ [ 1, $text{'edit_all'}."<br>",
+			"onClick='enable_cron_fields(\"$arr\", form, 0)'" ],
+		      [ 0, $text{'edit_selected'}."<br>",
+			"onClick='enable_cron_fields(\"$arr\", form, 1)'" ] ],
+		    $dis);
+	$col .= "<table> <tr>\n";
+        for(my $j=0; $j<@arrlist; $j+=($arr eq "mins" && $hourly_only ? 60 : 12)) {
+                my $jj = $j+($arr eq "mins" && $hourly_only ? 59 : 11);
+		if ($jj >= @arrlist) { $jj = @arrlist - 1; }
+		my @sec = @arrlist[$j .. $jj];
+		my @opts;
+		foreach my $v (@sec) {
+			if ($v =~ /^(.*)=(.*)$/) {
+				push(@opts, [ $2, $1 ]);
+				}
+			else {
+				push(@opts, [ $v, $v ]);
+				}
+			}
+		my $dis = $job->{$arr} eq "*" || $job->{$arr} eq "";
+		$col .= "<td valign=top>".
+			&ui_select($arr, [ keys %inuse ], \@opts,
+			  @sec > 12 ? ($arr eq "mins" && $hourly_only ? 1 : 12)
+                                  : scalar(@sec),
+			  $arr eq "mins" && $hourly_only ? 0 : 1,
+			  0, $dis).
+			"</td>\n";
+		}
+	$col .= "</tr></table>\n";
+	push(@cols, $col);
+	}
+$table .= &ui_columns_row(\@cols, [ "valign=top", "valign=top", "valign=top",
+				    "valign=top", "valign=top" ]);
+$table .= &ui_columns_end();
+$table .= &ui_note($text{'edit_ctrl'}, 0);
+$rv .= &ui_table_row(undef, $table, $width, undef, ['data-schedule-tr']);
+return $rv;
+}
+
 =head2 show_times_input(&job, [nospecial])
 
 Print HTML for inputs for selecting the schedule for a cron job, defined
@@ -764,7 +944,7 @@ return &theme_show_times_input(@_) if (defined(&theme_show_times_input));
 local $job = $_[0];
 if ($config{'vixie_cron'} && (!$_[1] || $_[0]->{'special'})) {
 	# Allow selection of special @ times
-	print "<tr $cb> <td colspan=6>\n";
+	print "<tr data-schedule-tr $cb> <td colspan=6>\n";
 	printf "<input type=radio name=special_def value=1 %s> %s\n",
 		$job->{'special'} ? "checked" : "", $text{'edit_special1'};
 	print "<select name=special onChange='change_special_mode(form, 1)'>\n";
@@ -772,7 +952,7 @@ if ($config{'vixie_cron'} && (!$_[1] || $_[0]->{'special'})) {
 	local $sp = $job->{'special'} eq 'midnight' ? 'daily' :
 	    $job->{'special'} eq 'annually' ? 'yearly' : $job->{'special'};
 	foreach $s ('hourly', 'daily', 'weekly', 'monthly', 'yearly', 'reboot'){
-		printf "<option value=%s %s>%s\n",
+		printf "<option value=%s %s>%s</option>\n",
 		    $s, $sp eq $s ? "selected" : "", $text{'edit_special_'.$s};
 		}
 	print "</select>\n";
@@ -796,8 +976,10 @@ change_special_mode(form, 0);
 
 function change_special_mode(form, special)
 {
-form.special_def[0].checked = special;
-form.special_def[1].checked = !special;
+  if(form.special_def) {
+    form.special_def[0].checked = special;
+    form.special_def[1].checked = !special;
+  }
 }
 </script>
 EOF
@@ -868,14 +1050,15 @@ foreach $arr ("mins", "hours", "days", "months", "weekdays") {
 		foreach $v (@sec) {
 			if ($v =~ /^(.*)=(.*)$/) { $disp = $1; $code = $2; }
 			else { $disp = $code = $v; }
-			printf "<option value=\"$code\" %s>$disp\n",
+			printf "<option value=\"$code\" %s>$disp</option>\n",
 				$inuse{$code} ? "selected" : "";
 			}
 		print "</select></td>\n";
 		}
 	print "</tr></table></td>\n";
 	}
-print "</tr> <tr $cb> <td colspan=5>$text{'edit_ctrl'}</td> </tr>\n";
+my $ctlnote = &ui_note($text{'edit_ctrl'}, 0);
+print "</tr> <tr $cb> <td colspan=5>$ctlnote</td> </tr>\n";
 }
 
 =head2 parse_times_input(&job, &in)
@@ -1028,62 +1211,6 @@ foreach $w (@wds) {
 $_[0]->{'weekdays'} = join(",", @wds);
 }
 
-=head2 create_wrapper(wrapper-path, module, script)
-
-Creates a wrapper script which calls a script in some module's directory
-with the proper webmin environment variables set. This should always be used
-when setting up a cron job, instead of attempting to run a command in the
-module directory directly.
-
-The parameters are :
-
-=item wrapper-path - Full path to the wrapper to create, like /etc/webmin/yourmodule/foo.pl
-
-=item module - Module containing the real script to call.
-
-=item script - Program within that module for the wrapper to run.
-
-=cut
-sub create_wrapper
-{
-local $perl_path = &get_perl_path();
-&open_tempfile(CMD, ">$_[0]");
-&print_tempfile(CMD, <<EOF
-#!$perl_path
-open(CONF, "$config_directory/miniserv.conf") || die "Failed to open $config_directory/miniserv.conf : \$!";
-while(<CONF>) {
-        \$root = \$1 if (/^root=(.*)/);
-        }
-close(CONF);
-\$root || die "No root= line found in $config_directory/miniserv.conf";
-\$ENV{'PERLLIB'} = "\$root";
-\$ENV{'WEBMIN_CONFIG'} = "$ENV{'WEBMIN_CONFIG'}";
-\$ENV{'WEBMIN_VAR'} = "$ENV{'WEBMIN_VAR'}";
-EOF
-	);
-if ($gconfig{'os_type'} eq 'windows') {
-	# On windows, we need to chdir to the drive first, and use system
-	&print_tempfile(CMD, "if (\$root =~ /^([a-z]:)/i) {\n");
-	&print_tempfile(CMD, "       chdir(\"\$1\");\n");
-	&print_tempfile(CMD, "       }\n");
-	&print_tempfile(CMD, "chdir(\"\$root/$_[1]\");\n");
-	&print_tempfile(CMD, "exit(system(\"\$root/$_[1]/$_[2]\", \@ARGV));\n");
-	}
-else {
-	# Can use exec on Unix systems
-	if ($_[1]) {
-		&print_tempfile(CMD, "chdir(\"\$root/$_[1]\");\n");
-		&print_tempfile(CMD, "exec(\"\$root/$_[1]/$_[2]\", \@ARGV) || die \"Failed to run \$root/$_[1]/$_[2] : \$!\";\n");
-		}
-	else {
-		&print_tempfile(CMD, "chdir(\"\$root\");\n");
-		&print_tempfile(CMD, "exec(\"\$root/$_[2]\", \@ARGV) || die \"Failed to run \$root/$_[2] : \$!\";\n");
-		}
-	}
-&close_tempfile(CMD);
-chmod(0755, $_[0]);
-}
-
 =head2 cron_file(&job)
 
 Returns the file that a cron job is in, or will be in when it is created
@@ -1103,28 +1230,53 @@ Returns a human-readable text string describing when a cron job is run.
 =cut
 sub when_text
 {
-local $pfx = $_[1] ? "uc" : "";
-if ($_[0]->{'special'}) {
-	$pfx = $_[1] ? "" : "lc";
-	return $text{$pfx.'edit_special_'.$_[0]->{'special'}};
+my ($job, $ucfirst) = @_;
+my $pfx = $ucfirst ? "uc" : "";
+if ($job->{'interval'}) {
+	return &text($pfx.'when_interval', $job->{'interval'});
 	}
-elsif ($_[0]->{'mins'} eq '*' && $_[0]->{'hours'} eq '*' && $_[0]->{'days'} eq '*' && $_[0]->{'months'} eq '*' && $_[0]->{'weekdays'} eq '*') {
+elsif ($job->{'special'}) {
+	$pfx = $_[1] ? "" : "lc";
+	return $text{$pfx.'edit_special_'.$job->{'special'}};
+	}
+elsif ($job->{'boot'}) {
+	return &text($pfx.'when_boot');
+	}
+elsif ($job->{'mins'} eq '*' && $job->{'hours'} eq '*' &&
+       $job->{'days'} eq '*' && $job->{'months'} eq '*' &&
+       $job->{'weekdays'} eq '*') {
 	return $text{$pfx.'when_min'};
 	}
-elsif ($_[0]->{'mins'} =~ /^\d+$/ && $_[0]->{'hours'} eq '*' && $_[0]->{'days'} eq '*' && $_[0]->{'months'} eq '*' && $_[0]->{'weekdays'} eq '*') {
-	return &text($pfx.'when_hour', $_[0]->{'mins'});
+elsif ($job->{'mins'} =~ /^\d+$/ && $job->{'hours'} eq '*' &&
+       $job->{'days'} eq '*' && $job->{'months'} eq '*' &&
+       $job->{'weekdays'} eq '*') {
+	return &text($pfx.'when_hour', $job->{'mins'});
 	}
-elsif ($_[0]->{'mins'} =~ /^\d+$/ && $_[0]->{'hours'} =~ /^\d+$/ && $_[0]->{'days'} eq '*' && $_[0]->{'months'} eq '*' && $_[0]->{'weekdays'} eq '*') {
-	return &text($pfx.'when_day', sprintf("%2.2d", $_[0]->{'mins'}), $_[0]->{'hours'});
+elsif ($job->{'mins'} =~ /^\d+$/ && $job->{'hours'} =~ /^\d+$/ &&
+       $job->{'days'} eq '*' && $job->{'months'} eq '*' &&
+       $job->{'weekdays'} eq '*') {
+	return &text($pfx.'when_day',
+		     sprintf("%2.2d", $job->{'mins'}),
+		     sprintf("%2.2d", $job->{'hours'}));
 	}
-elsif ($_[0]->{'mins'} =~ /^\d+$/ && $_[0]->{'hours'} =~ /^\d+$/ && $_[0]->{'days'} =~ /^\d+$/ && $_[0]->{'months'} eq '*' && $_[0]->{'weekdays'} eq '*') {
-	return &text($pfx.'when_month', sprintf("%2.2d", $_[0]->{'mins'}), $_[0]->{'hours'}, $_[0]->{'days'});
+elsif ($job->{'mins'} =~ /^\d+$/ && $job->{'hours'} =~ /^\d+$/ &&
+       $job->{'days'} =~ /^\d+$/ && $job->{'months'} eq '*' &&
+       $job->{'weekdays'} eq '*') {
+	return &text($pfx.'when_month',
+		     sprintf("%2.2d", $job->{'mins'}),
+		     sprintf("%2.2d", $job->{'hours'}),
+		     $job->{'days'});
 	}
-elsif ($_[0]->{'mins'} =~ /^\d+$/ && $_[0]->{'hours'} =~ /^\d+$/ && $_[0]->{'days'} eq '*' && $_[0]->{'months'} eq '*' && $_[0]->{'weekdays'} =~ /^\d+$/) {
-	return &text($pfx.'when_weekday', sprintf("%2.2d", $_[0]->{'mins'}), $_[0]->{'hours'}, $text{"day_".$_[0]->{'weekdays'}});
+elsif ($job->{'mins'} =~ /^\d+$/ && $job->{'hours'} =~ /^\d+$/ &&
+       $job->{'days'} eq '*' && $job->{'months'} eq '*' &&
+       $job->{'weekdays'} =~ /^\d+$/) {
+	return &text($pfx.'when_weekday',
+		     sprintf("%2.2d", $job->{'mins'}),
+		     sprintf("%2.2d", $job->{'hours'}),
+		     $text{"day_".$job->{'weekdays'}});
 	}
 else {
-	return &text($pfx.'when_cron', join(" ", $_[0]->{'mins'}, $_[0]->{'hours'}, $_[0]->{'days'}, $_[0]->{'months'}, $_[0]->{'weekdays'}));
+	return &text($pfx.'when_cron', join(" ", $job->{'mins'}, $job->{'hours'}, $job->{'days'}, $job->{'months'}, $job->{'weekdays'}));
 	}
 }
 
@@ -1136,15 +1288,17 @@ cron.deny files.
 =cut
 sub can_use_cron
 {
+local ($user) = @_;
+defined(getpwnam($user)) || return 0;	# User does not exist
 local $err;
 if (-r $config{cron_allow_file}) {
 	local @allowed = &list_allowed();
-	if (&indexof($_[0], @allowed) < 0 &&
+	if (&indexof($user, @allowed) < 0 &&
 	    &indexof("all", @allowed) < 0) { $err = 1; }
 	}
 elsif (-r $config{cron_deny_file}) {
 	local @denied = &list_denied();
-	if (&indexof($_[0], @denied) >= 0 ||
+	if (&indexof($user, @denied) >= 0 ||
 	    &indexof("all", @denied) >= 0) { $err = 1; }
 	}
 elsif ($config{cron_deny_all} == 0) { $err = 1; }
@@ -1205,10 +1359,31 @@ $cmd =~ s/^\s*\[.*\]\s+\|\|\s+//;
 while($cmd =~ s/(\d*)(<|>)((\/\S+)|&\d+)\s*$//) { }
 $cmd =~ s/^\((.*)\)\s*$/$1/;
 $cmd =~ s/\s+$//;
+my $eos;
 if ($config{'match_mode'} == 1) {
 	$cmd =~ s/\s.*$//;
+	$eos = '$';
 	}
-($proc) = grep { $_->{'args'} =~ /\Q$cmd\E/ &&
+else {
+	my $cmd_ = $cmd;
+	$cmd_ =~ s/\s.*$//;
+	if ($cmd_ eq $cmd) {
+		$eos = '$';
+		}
+	}
+
+# If `Input to command` is set, remove it for test case
+# otherwise cmd will always be displayed as not running
+my $cmd_ = $cmd;
+
+# First remove only input to command and preserve
+# potential arguments containing escaped %, like \\%
+$cmd_ =~ s/(?<!\\)%.*//;
+
+# Now replace escaped \\ special chars for testing purposes
+$cmd_ =~ s/\\//g;
+
+($proc) = grep { $_->{'args'} =~ /\Q$cmd_\E$eos/ &&
 		 (!$config{'match_user'} || $_->{'user'} eq $_[0]->{'user'}) }
 		@procs;
 if (!$proc && $cmd =~ /^$config_directory\/(.*\.pl)(.*)$/) {
@@ -1250,8 +1425,8 @@ sub extract_input
 local ($cmd) = @_;
 $cmd =~ s/\\%/\0/g;
 local ($cmd, $input) = split(/\%/, $cmd, 2);
-$cmd =~ s/\0/%/g;
-$input =~ s/\0/%/g;
+$cmd =~ s/\0/\\%/g;
+$input =~ s/\0/\\%/g;
 return ($cmd, $input);
 }
 
@@ -1314,7 +1489,7 @@ Given a cron job with a # comment after the command, sets the comment field
 sub convert_comment
 {
 local ($job) = @_;
-if ($job->{'command'} =~ /^(.*)\s*#([^#]*)$/) {
+if ($job->{'command'} =~ /^(.*\S)\s*#([^#]*)$/) {
 	$job->{'command'} = $1;
 	$job->{'comment'} = $2;
 	return 1;
@@ -1350,14 +1525,20 @@ sub check_cron_config
 if ($config{'single_file'} && !-r $config{'single_file'}) {
 	return &text('index_esingle', "<tt>$config{'single_file'}</tt>");
 	}
-if ($config{'cron_get_command'} =~ /^(\S+)/ && !&has_command("$1")) {
+if (!&has_crontab_cmd() && $config{'cron_get_command'} =~ /^(\S+)/ &&
+    !&has_command("$1")) {
 	return &text('index_ecmd', "<tt>$1</tt>");
 	}
 # Check for directory
 local $fcron = ($config{'cron_dir'} =~ /\/fcron$/);
 if (!$single_user && !$config{'single_file'} &&
     !$fcron && !-d $config{'cron_dir'}) {
-	return &text('index_ecrondir', "<tt>$config{'cron_dir'}</tt>");
+	if (!$in{'create_dir'}) {
+		return &text('index_ecrondir', "<tt>$config{'cron_dir'}</tt>").
+		"<p><a href=\"index.cgi?create_dir=yes\">".&text('index_ecrondir_create' ,"<tt>$config{'cron_dir'}</tt>")."</a></p>";
+	} else {
+		&make_dir($config{'cron_dir'}, 0755);
+		}
 	}
 return undef;
 }
@@ -1377,7 +1558,8 @@ if ($err) {
 
 =head2 cleanup_temp_files
 
-Called from cron to delete old files in the Webmin /tmp directory
+Called from cron to delete old files in the Webmin /tmp directory, and also
+old lock links directories.
 
 =cut
 sub cleanup_temp_files
@@ -1387,27 +1569,221 @@ if (!$gconfig{'tempdelete_days'}) {
 	print STDERR "Temp file clearing is disabled\n";
 	return;
 	}
+
+# Cleanup files in /tmp/.webmin
 if ($gconfig{'tempdir'} && !$gconfig{'tempdirdelete'}) {
 	print STDERR "Temp file clearing is not done for the custom directory $gconfig{'tempdir'}\n";
-	return;
 	}
+else {
+	my $tempdir = &transname();
+	$tempdir =~ s/\/([^\/]+)$//;
+	if (!$tempdir || $tempdir eq "/") {
+		$tempdir = "/tmp/.webmin";
+		}
 
-local $tempdir = &transname();
-$tempdir =~ s/\/([^\/]+)$//;
-if (!$tempdir || $tempdir eq "/") {
-	$tempdir = "/tmp/.webmin";
-	}
-
-local $cutoff = time() - $gconfig{'tempdelete_days'}*24*60*60;
-opendir(DIR, $tempdir);
-foreach my $f (readdir(DIR)) {
-	next if ($f eq "." || $f eq "..");
-	local @st = lstat("$tempdir/$f");
-	if ($st[9] < $cutoff) {
-		&unlink_file("$tempdir/$f");
+	my $cutoff = time() - $gconfig{'tempdelete_days'}*24*60*60;
+	if (opendir(DIR, $tempdir)) {
+		foreach my $f (readdir(DIR)) {
+			next if ($f eq "." || $f eq "..");
+			my @st = lstat("$tempdir/$f");
+			if ($st[9] < $cutoff) {
+				&unlink_file("$tempdir/$f");
+				}
+			}
+		closedir(DIR);
 		}
 	}
-closedir(DIR);
+
+# Delete stale lock files
+my $lockdir = $var_directory."/locks";
+if (opendir(DIR, $lockdir)) {
+	foreach my $f (readdir(DIR)) {
+		next if ($f eq "." || $f eq "..");
+		next if ($f !~ /^\d+$/);
+		if (!kill(0, $f)) {
+			# Process is gone, but never cleaned up!
+			&unlink_file("$lockdir/$f");
+			}
+		}
+	closedir(DIR);
+	}
+
+# Cleanup old websockets
+foreach (&get_miniserv_websockets_modules()) {
+	&cleanup_miniserv_websockets(undef, $_);
+	}
+
+# Delete forgot-password files older than 1 day
+if (opendir(DIR, $main::forgot_password_link_dir)) {
+	my $cutoff = time() - 24*60*60;
+	foreach my $f (readdir(DIR)) {
+		next if ($f eq "." || $f eq "..");
+		next if ($f eq "ratelimit");
+		my $path = $main::forgot_password_link_dir."/".$f;
+		my @st = stat($path);
+		if ($st[9] < $cutoff) {
+			&unlink_file($path);
+			}
+		}
+	closedir(DIR);
+	}
+}
+
+=head2 list_cron_files()
+
+Returns a list of all files containing cron jobs
+
+=cut
+sub list_cron_files
+{
+my @jobs = &list_cron_jobs();
+my @files = map { $_->{'file'} } grep { $_->{'file'} } @jobs;
+if ($config{'system_crontab'}) {
+	push(@files, $config{'system_crontab'});
+	}
+if ($config{'cronfiles_dir'}) {
+	push(@files, glob(&translate_filename($config{'cronfiles_dir'})."/*"));
+	@files = grep { -f $_ } @files;
+	}
+return &unique(@files);
+}
+
+=head2 has_crontab_cmd()
+
+Returns 1 if the crontab command exists on this system
+
+=cut
+sub has_crontab_cmd
+{
+my $cmd = $config{'cron_edit_command'};
+if ($cmd) {
+	$cmd =~ s/^su.*-c\s+//;
+	($cmd) = &split_quoted_string($cmd);
+	my $rv = &has_command($cmd);
+	return $rv if ($rv);
+	}
+return &has_command("crontab");
+}
+
+=head2 next_run(&job)
+
+Given a cron job, returns the unix time on which it will run next.
+
+=cut
+sub next_run
+{
+my ($job) = @_;
+my $now = time();
+my @tm = localtime($now);
+if ($job->{'special'} eq 'hourly') {
+	$job = { 'mins' => 0,
+		 'hours' => '*',
+		 'days' => '*',
+		 'months' => '*',
+		 'weekdays' => '*' };
+	}
+elsif ($job->{'special'} eq 'daily') {
+	$job = { 'mins' => 0,
+		 'hours' => 0,
+		 'days' => '*',
+		 'months' => '*',
+		 'weekdays' => '*' };
+	}
+elsif ($job->{'special'} eq 'weekly') {
+	$job = { 'mins' => 0,
+		 'hours' => 0,
+		 'days' => '*',
+		 'months' => '*',
+		 'weekdays' => 0 };
+	}
+elsif ($job->{'special'} eq 'monthly') {
+	$job = { 'mins' => 0,
+		 'hours' => 0,
+		 'days' => '1',
+		 'months' => '*',
+		 'weekdays' => '*' };
+	}
+elsif ($job->{'special'} eq 'yearly') {
+	$job = { 'mins' => 0,
+		 'hours' => 0,
+		 'days' => 1,
+		 'months' => 1,
+		 'weekdays' => '*' };
+	}
+elsif ($job->{'special'} eq 'reboot') {
+	return undef;
+	}
+my @mins = &cron_all_ranges($job->{'mins'}, 0, 59);
+my @hours = &cron_all_ranges($job->{'hours'}, 0, 23);
+my @days = &cron_all_ranges($job->{'days'}, 1, 31);
+my @months = &cron_all_ranges($job->{'months'}, 1, 12);
+my @weekdays = &cron_all_ranges($job->{'weekdays'}, 0, 6);
+my ($min, $hour, $day, $month, $year);
+my @possible;
+foreach $min (@mins) {
+	foreach $hour (@hours) {
+		foreach $day (@days) {
+			foreach $month (@months) {
+				foreach $year ($tm[5] .. $tm[5]+7) {
+					my $tt;
+					eval { $tt = timelocal(0, $min, $hour, $day, $month-1, $year) };
+					next if ($tt < $now);
+					my @ttm = localtime($tt);
+					next if (&indexof($ttm[6], @weekdays) < 0);
+					push(@possible, $tt);
+					last;
+					}
+				}
+			}
+		}
+	}
+@possible = sort { $a <=> $b } @possible;
+return $possible[0];
+}
+
+=head2 cron_range(range, min, max)
+
+=cut
+sub cron_range
+{
+my ($w, $min, $max) = @_;
+my $j;
+my %inuse;
+if ($w eq "*") {
+	# all values
+	for($j=$min; $j<=$max; $j++) { $inuse{$j}++; }
+	}
+elsif ($w =~ /^\*\/(\d+)$/) {
+	# only every Nth
+	my $step = $1 || 1;
+	for($j=$min; $j<=$max; $j+=$step) { $inuse{$j}++; }
+	}
+elsif ($w =~ /^(\d+)-(\d+)\/(\d+)$/) {
+	# only every Nth of some range
+	my $step = $3 || 1;
+	for($j=$1; $j<=$2; $j+=$step) { $inuse{int($j)}++; }
+	}
+elsif ($w =~ /^(\d+)-(\d+)$/) {
+	# all of some range
+	for($j=$1; $j<=$2; $j++) { $inuse{int($j)}++; }
+	}
+else {
+	# One value
+	$inuse{int($w)}++;
+	}
+return sort { $a <=> $b } (keys %inuse);
+}
+
+=head2 cron_all_ranges(comma-list, min, max)
+
+=cut
+sub cron_all_ranges
+{
+my @rv;
+foreach $r (split(/,/, $_[0])) {
+	push(@rv, &cron_range($r, $_[1], $_[2]));
+	}
+return sort { $a <=> $b } @rv;
 }
 
 1;
