@@ -13,7 +13,7 @@ sub list_mounts
 local(@rv, @p, @o, $_, $i, $j); $i = 0;
 
 # Get /etc/fstab mounts
-open(FSTAB, $config{'fstab_file'});
+open(FSTAB, "<".$config{'fstab_file'});
 while(<FSTAB>) {
 	local(@o, $at_boot);
 	chop; s/#.*$//g;
@@ -25,7 +25,7 @@ while(<FSTAB>) {
 	$rv[$i]->[5] = "yes";
 	@o = split(/,/ , $p[3] eq "defaults" ? "" : $p[3]);
 	if (($j = &indexof("noauto", @o)) >= 0) {
-		# filesytem is not mounted at boot
+		# filesystem is not mounted at boot
 		splice(@o, $j, 1);
 		$rv[$i]->[5] = "no";
 		}
@@ -67,7 +67,7 @@ local($i, @fstab, $line, $opts, $j, @amd);
 $i = 0;
 
 # Update fstab file
-open(FSTAB, $config{'fstab_file'});
+open(FSTAB, "<".$config{'fstab_file'});
 @fstab = <FSTAB>;
 close(FSTAB);
 &open_tempfile(FSTAB, ">$config{'fstab_file'}");
@@ -101,7 +101,7 @@ local($i, @fstab, $line, $opts, $j, @amd);
 $i = 0;
 
 # Update fstab file
-open(FSTAB, $config{'fstab_file'});
+open(FSTAB, "<".$config{'fstab_file'});
 @fstab = <FSTAB>;
 close(FSTAB);
 &open_tempfile(FSTAB, ">$config{'fstab_file'}");
@@ -128,8 +128,9 @@ sub list_mounted
 {
 # get the list of mounted filesystems
 local(@rv, $_);
-local $cmd = $uname_release =~ /^2\.[0-9]/ ? "netbsd-mounts-2"
-					   : "netbsd-mounts";
+local $cmd = $uname_release =~ /^1\.[0-9]/ ? "netbsd-mounts" :
+	     $uname_release =~ /^2\.[0-9]/ ? "netbsd-mounts-2" :
+					     "netbsd-mounts-3";
 &compile_program($cmd);
 open(CMD, "$module_config_directory/$cmd |");
 while(<CMD>) {
@@ -230,9 +231,9 @@ sub disk_space
 {
 if (&get_mounted($_[1], "*") < 0) { return (); }
 if ($_[0] eq "proc" || $_[0] eq "swap") { return (); }
-if (&backquote_command("df -k $_[1]", 1) =~
-    /Mounted on\n\S+\s+(\S+)\s+\S+\s+(\S+)/) {
-	return ($1, $2);
+my $out = &backquote_command("df -k ".quotemeta($_[1]), 1);
+if ($out =~ /Mounted on\n\S+\s+(?<total>\S+)\s+(?<used>\S+)\s+(?<free>\S+)\s+(?<percent>\d+)/) {
+	return ("$+{total}", "$+{free}", "$+{used}", "$+{percent}");
 	}
 return ( );
 }
@@ -278,54 +279,57 @@ return $_[0] eq "nfs";
 # Output HTML for editing the mount location of some filesystem.
 sub generate_location
 {
-if ($_[0] eq "nfs") {
+local ($type, $loc) = @_;
+if ($type eq "nfs") {
 	# NFS mount from some host and directory
-	$_[1] =~ /^([^:]+):(.*)$/;
-	print "<tr> <td><b>NFS Hostname</b></td>\n";
-	print "<td><input name=nfs_host size=20 value=\"$1\">\n";
-	&nfs_server_chooser_button("nfs_host");
-	print "</td>\n";
-	print "<td><b>NFS Directory</b></td>\n";
-	print "<td><input name=nfs_dir size=20 value=\"$2\">\n";
-	&nfs_export_chooser_button("nfs_host", "nfs_dir");
-	print "</td> </tr>\n";
+        local ($host, $dir) = $loc =~ /^([^:]+):(.*)$/ ? ( $1, $2 ) : ( );
+        print &ui_table_row(&hlink($text{'linux_nfshost'}, "nfshost"),
+                &ui_textbox("nfs_host", $host, 30).
+                &nfs_server_chooser_button("nfs_host").
+                "&nbsp;".
+                "<b>".&hlink($text{'linux_nfsdir'}, "nfsdir")."</b> ".
+                &ui_textbox("nfs_dir",
+                            ($type eq "nfs4") && ($dir eq "") ? "/" : $dir, 30).
+                &nfs_export_chooser_button("nfs_host", "nfs_dir"));
 	}
 else {
-	if ($_[0] eq "swap") {
-		# Swap file or device
-		printf "<tr> <td valign=top><b>Swap File</b></td>\n";
-		}
-	else {
-		# Disk-based filesystem
-		printf "<tr> <td valign=top><b>%s Disk</b></td>\n",
-			&fstype_name($_[0]);
-		}
-	print "<td colspan=3>\n";
-	if ($_[1] =~ /^\/dev\/wd(\d)s(\d)([a-z]*)$/) {
+	local $msg;
+        if ($type eq "swap") {
+                # Swap file or device
+		$msg = $text{'linux_swapfile'};
+                }
+        else {
+                # Disk-based filesystem
+                $msg = &fstype_name($type);
+                }
+	local ($disk_dev, $ide_t, $ide_s, $ide_p, $scsi_t, $scsi_s, $scsi_p);
+	if ($loc =~ /^\/dev\/wd(\d)s(\d)([a-z]*)$/) {
 		$disk_dev = 0; $ide_t = $1; $ide_s = $2; $ide_p = $3;
 		}
-	elsif ($_[1] =~ /^\/dev\/sd(\d)s(\d)([a-z]*)$/) {
+	elsif ($loc =~ /^\/dev\/sd(\d)s(\d)([a-z]*)$/) {
 		$disk_dev = 1; $scsi_t = $1; $scsi_s = $2; $scsi_p = $3;
 		}
 	else { $disk_dev = 2; }
 
-	printf "<input type=radio name=disk_dev value=0 %s> IDE Hard Disk:\n",
-		$disk_dev == 0 ? "checked" : "";
-	print "Device <input name=ide_t size=3 value=\"$ide_t\">\n";
-	print "Slice <input name=ide_s size=3 value=\"$ide_s\">\n";
-	print "Partition <input name=ide_p size=3 value=\"$ide_p\"><br>\n";
-
-	printf "<input type=radio name=disk_dev value=1 %s> SCSI Disk:\n",
-		$disk_dev == 1 ? "checked" : "";
-	print "Device <input name=scsi_t size=3 value=\"$scsi_t\">\n";
-	print "Slice <input name=scsi_s size=3 value=\"$scsi_s\">\n";
-	print "Partition <input name=scsi_p size=3 value=\"$scsi_p\"><br>\n";
-
-	printf "<input type=radio name=disk_dev value=2 %s> Other Device:\n",
-		$disk_dev == 2 ? "checked" : "";
-	printf "<input size=20 name=dev_path value=\"%s\"><br>\n",
-		$disk_dev == 2 ? $_[1] : "";
-	print "</td> </tr>\n";
+        print &ui_table_row($msg,
+                &ui_radio_table("disk_dev", $disk_dev,
+                  [ [ 0, $text{'freebsd_ide'},
+                      $text{'freebsd_device'}." ".
+                        &ui_textbox("ide_t", $ide_t, 4)." ".
+                      $text{'freebsd_slice'}." ".
+                        &ui_textbox("ide_s", $ide_s, 4)." ".
+                      $text{'freebsd_part'}." ".
+                        &ui_textbox("ide_p", $ide_p, 4) ],
+                    [ 1, $text{'freebsd_scsi'},
+                      $text{'freebsd_device'}." ".
+                        &ui_textbox("scsi_t", $scsi_t, 4)." ".
+                      $text{'freebsd_slice'}." ".
+                        &ui_textbox("scsi_s", $scsi_s, 4)." ".
+                      $text{'freebsd_part'}." ".
+                        &ui_textbox("scsi_p", $scsi_p, 4) ],
+                    [ 2, $text{'freebsd_other'},
+                      &ui_textbox("dev_path", $disk_dev == 2 ? $loc : "", 40).
+                      " ".&file_chooser_button("dev_path", 0) ] ]));
 	}
 }
 
@@ -516,25 +520,28 @@ elsif ($_[0] eq "swap") {
 sub check_location
 {
 if ($_[0] eq "nfs") {
-	local($out, $temp, $mout, $dirlist);
+	local($out, $temp, $mout, $dirlist, $qnfs_host, $qnfs_dir, $qtemp);
+	$qnfs_host = quotemeta($in{nfs_host});
 
-	# Use ping and showmount to see if the host exists and is up
-	if ($in{nfs_host} !~ /^\S+$/) {
-		&error("'$in{nfs_host}' is not a valid hostname");
-		}
-	$out = &backquote_command("ping -c 1 '$in{nfs_host}' 2>&1");
-	if ($out =~ /unknown host/i) {
-		&error("The host '$in{nfs_host}' does not exist");
-		}
-	elsif ($out =~ /100\% packet loss/) {
-		&error("The host '$in{nfs_host}' is down");
-		}
-	$out = &backquote_command("showmount -e '$in{nfs_host}' 2>&1");
-	if ($out =~ /Unable to receive/) {
-		&error("The host '$in{nfs_host}' does not support NFS");
-		}
-	elsif ($?) {
-		&error("Failed to get mount list : $out");
+	if ($config{'nfs_check'}) {
+		# Use ping and showmount to see if the host exists and is up
+		if ($in{nfs_host} !~ /^\S+$/) {
+			&error("'$in{nfs_host}' is not a valid hostname");
+			}
+		$out = &backquote_command("ping -c 1 $qnfs_host 2>&1");
+		if ($out =~ /unknown host/i) {
+			&error("The host '$in{nfs_host}' does not exist");
+			}
+		elsif ($out =~ /100\% packet loss/) {
+			&error("The host '$in{nfs_host}' is down");
+			}
+		$out = &backquote_command("showmount -e $qnfs_host 2>&1");
+		if ($out =~ /Unable to receive/) {
+			&error("The host '$in{nfs_host}' does not support NFS");
+			}
+		elsif ($?) {
+			&error("Failed to get mount list : $out");
+			}
 		}
 
 	# Validate directory name
@@ -550,7 +557,9 @@ if ($_[0] eq "nfs") {
 	# Try a test mount to see if filesystem is available
 	$temp = &transname();
 	&make_dir($temp, 0755);
-	$mout = &backquote_command("mount $in{nfs_host}:$in{nfs_dir} $temp 2>&1");
+	$qnfs_dir = quotemeta($in{nfs_dir});
+	$qtemp = quotemeta($temp);
+	$mout = &backquote_command("mount $qnfs_host:$qnfs_dir $qtemp 2>&1");
 	if ($mout =~ /No such file or directory/) {
 		&error("The directory '$in{nfs_dir}' does not exist on the ".
 		       "host $in{nfs_host}. The available directories are:".
@@ -718,12 +727,14 @@ return @rv ? join(',' , @rv) : "-";
 # Calls dd and mkswap to setup a swap file
 sub create_swap
 {
-local($out, $bl);
-$bl = $_[1] * ($_[2] eq "g" ? 1024*1024 :
+local($out, $bl, $qfile);
+$qfile = quotemeta($_[0]);
+$bl = $_[1] * ($_[2] eq "t" ? 1024*1024*1024 :
+	       $_[2] eq "g" ? 1024*1024 :
 	       $_[2] eq "m" ? 1024 : 1);
-$out = &backquote_logged("dd if=/dev/zero of=$_[0] bs=1024 count=$bl 2>&1");
+$out = &backquote_logged("dd if=/dev/zero of=$qfile bs=1024 count=$bl 2>&1");
 if ($?) { return "dd failed : $out"; }
-$out = &backquote_logged("mkswap $_[0] $bl 2>&1");
+$out = &backquote_logged("mkswap $qfile $bl 2>&1");
 if ($?) { return "mkswap failed : $out"; }
 &system_logged("sync >/dev/null 2>&1");
 return 0;
